@@ -24,6 +24,11 @@
 
 #include <inttypes.h>
 #include <asm/unistd.h>
+#include <linux/audit.h>
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /*
  * types
@@ -79,6 +84,21 @@ struct scmp_arg_cmp {
  * macros/defines
  */
 
+/**
+ * The native architecture token
+ */
+#define SCMP_ARCH_NATIVE	0
+
+/**
+ * The x86 (32-bit) architecture token
+ */
+#define SCMP_ARCH_X86		AUDIT_ARCH_I386
+
+/**
+ * The x86-64 (64-bit) architecture token
+ */
+#define SCMP_ARCH_X86_64	AUDIT_ARCH_X86_64
+
 /**
  * Convert a syscall name into the associated syscall number
  * @param x the syscall name
@@ -165,7 +185,7 @@ struct scmp_arg_cmp {
 scmp_filter_ctx seccomp_init(uint32_t def_action);
 
 /**
- * Reset the current filter state
+ * Reset the filter state
  * @param ctx the filter context
  * @param def_action the default filter action
  *
@@ -178,7 +198,7 @@ scmp_filter_ctx seccomp_init(uint32_t def_action);
 int seccomp_reset(scmp_filter_ctx ctx, uint32_t def_action);
 
 /**
- * Destroys the current filter state and releases any resources
+ * Destroys the filter state and releases any resources
  * @param ctx the filter context
  *
  * This functions destroys the given seccomp filter state and releases any
@@ -190,7 +210,7 @@ int seccomp_reset(scmp_filter_ctx ctx, uint32_t def_action);
 void seccomp_release(scmp_filter_ctx ctx);
 
 /**
- * Loads the current filter into the kernel
+ * Loads the filter into the kernel
  * @param ctx the filter context
  *
  * This function loads the given seccomp filter context into the kernel.  If
@@ -243,7 +263,7 @@ int seccomp_syscall_priority(scmp_filter_ctx ctx,
 			     int syscall, uint8_t priority);
 
 /**
- * Add a new rule to the current filter
+ * Add a new rule to the filter
  * @param ctx the filter context
  * @param action the filter action
  * @param syscall the syscall number
@@ -262,7 +282,7 @@ int seccomp_rule_add(scmp_filter_ctx ctx,
 		     uint32_t action, int syscall, unsigned int arg_cnt, ...);
 
 /**
- * Add a new rule to the current filter
+ * Add a new rule to the filter
  * @param ctx the filter context
  * @param action the filter action
  * @param syscall the syscall number
@@ -306,6 +326,7 @@ int seccomp_export_bpf(const scmp_filter_ctx ctx, int fd);
  */
 
 /* NOTE - pseudo syscall values {-1..-99} are reserved */
+#define __NR_SCMP_ERROR		-1
 
 #define __PNR_socket		-100
 #ifndef __NR_socket
@@ -458,4 +479,8 @@ int seccomp_export_bpf(const scmp_filter_ctx ctx, int fd);
 #define __NR_shmctl		__PNR_shmctl
 #endif /* __NR_shmctl */
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
