@@ -310,6 +310,297 @@ int db_action_valid(uint32_t action)
 	return -EINVAL;
 }
 
+/**
+ * Free and reset the seccomp filter collection
+ * @param col the seccomp filter collection
+ * @param def_action the default filter action
+ *
+ * This function frees any existing filter DBs and resets the collection to a
+ * default state.
+ *
+ */
+void db_col_reset(struct db_filter_col *col, uint32_t def_action)
+{
+	unsigned int iter;
+
+	if (col == NULL)
+		return;
+
+	/* free any filters */
+	for (iter = 0; iter < col->filter_cnt; iter++)
+		db_release(col->filters[iter]);
+	col->filter_cnt = 0;
+	free(col->filters);
+	col->filters = NULL;
+
+	/* set the default attribute values */
+	col->attr.act_default = def_action;
+	col->attr.act_badarch = SCMP_ACT_KILL;
+	col->attr.nnp_enable = 1;
+
+	/* set the state */
+	col->state = _DB_STA_VALID;
+}
+
+/**
+ * Intitalize a seccomp filter collection
+ * @param def_action the default filter action
+ *
+ * This function initializes a seccomp filter collection and readies it for
+ * use.  Returns a pointer to the collection on success, NULL on failure.
+ *
+ */
+struct db_filter_col *db_col_init(uint32_t def_action)
+{
+	struct db_filter_col *col;
+
+	col = malloc(sizeof(*col));
+	if (col == NULL)
+		return NULL;
+
+	/* clear the buffer for the first time */
+	memset(col, 0, sizeof(*col));
+
+	/* reset the DB to a known state */
+	db_col_reset(col, def_action);
+
+	return col;
+}
+
+/**
+ * Destroy a seccomp filter collection
+ * @param col the seccomp filter collection
+ *
+ * This function destroys a seccomp filter collection.  After calling this
+ * function, the filter should no longer be referenced.
+ *
+ */
+void db_col_release(struct db_filter_col *col)
+{
+	if (col == NULL)
+		return;
+
+	/* set the state, just in case */
+	col->state = _DB_STA_FREED;
+
+	/* free and reset the DB */
+	db_col_reset(col, 0);
+	free(col);
+}
+
+/**
+ * Validate a filter collection
+ * @param col the seccomp filter collection
+ *
+ * This function validates a seccomp filter collection.  Returns zero if the
+ * collection is valid, negative values on failure.
+ *
+ */
+int db_col_valid(struct db_filter_col *col)
+{
+	if (col != NULL && col->state == _DB_STA_VALID)
+		return 0;
+	return -EINVAL;
+}
+
+/**
+ * Merge two filter collections
+ * @param col_dst the destination filter collection
+ * @param col_src the source filter collection
+ *
+ * This function merges two filter collections into the given destination
+ * collection.  The source filter collection is no longer valid if the function
+ * returns successfully.  Returns zero on success, negative values on failure.
+ *
+ */
+int db_col_merge(struct db_filter_col *col_dst, struct db_filter_col *col_src)
+{
+	unsigned int iter_a, iter_b;
+	struct db_filter **dbs;
+
+	/* make sure we don't have any arch/filter collisions */
+	for (iter_a = 0; iter_a < col_dst->filter_cnt; iter_a++) {
+		for (iter_b = 0; iter_b < col_src->filter_cnt; iter_b++) {
+			if (col_dst->filters[iter_a]->arch->token ==
+			    col_src->filters[iter_b]->arch->token)
+				return -EEXIST;
+		}
+	}
+
+	/* expand the destination */
+	dbs = realloc(col_dst->filters,
+		      sizeof(struct db_filter *) *
+		      (col_dst->filter_cnt + col_src->filter_cnt));
+	if (dbs == NULL)
+		return -ENOMEM;
+
+	/* transfer the architecture filters */
+	for (iter_a = col_dst->filter_cnt, iter_b = 0;
+	     iter_b < col_src->filter_cnt; iter_a++, iter_b++) {
+		col_dst->filters[iter_a] = col_src->filters[iter_b];
+		col_dst->filter_cnt++;
+	}
+
+	/* free the source */
+	col_src->filter_cnt = 0;
+	db_col_release(col_src);
+
+	return 0;
+}
+
+/**
+ * Check to see if an architecture filter exists in the filter collection
+ * @param col the seccomp filter collection
+ * @param arch_token the architecture token
+ *
+ * Iterate through the given filter collection checking to see if a filter
+ * exists for the specified architecture.  Returns -EEXIST if a filter is found,
+ * zero if a matching filter does not exist.
+ *
+ */
+int db_col_arch_exist(struct db_filter_col *col, uint32_t arch_token)
+{
+	unsigned int iter;
+
+	for (iter = 0; iter < col->filter_cnt; iter++)
+		if (col->filters[iter]->arch->token == arch_token)
+			return -EEXIST;
+
+	return 0;
+}
+
+/**
+ * Get a filter attribute
+ * @param col the seccomp filter collection
+ * @param attr the filter attribute
+ * @param value the filter attribute value
+ *
+ * Get the requested filter attribute and provide it via @value.  Returns zero
+ * on success, negative values on failure.
+ *
+ */
+int db_col_attr_get(const struct db_filter_col *col,
+		    enum scmp_filter_attr attr, uint32_t *value)
+{
+	switch (attr) {
+	case SCMP_FLTATR_ACT_DEFAULT:
+		*value = col->attr.act_default;
+		break;
+	case SCMP_FLTATR_ACT_BADARCH:
+		*value = col->attr.act_badarch;
+		break;
+	case SCMP_FLTATR_CTL_NNP:
+		*value = col->attr.nnp_enable;
+		break;
+	default:
+		return -EEXIST;
+		break;
+	}
+
+	return 0;
+}
+
+/**
+ * Set a filter attribute
+ * @param db the seccomp filter collection
+ * @param attr the filter attribute
+ * @param value the filter attribute value
+ *
+ * Set the requested filter attribute with the given value.  Returns zero on
+ * success, negative values on failure.
+ *
+ */
+int db_col_attr_set(struct db_filter_col *col,
+		    enum scmp_filter_attr attr, uint32_t value)
+{
+	switch (attr) {
+	case SCMP_FLTATR_ACT_DEFAULT:
+		/* read only */
+		return -EACCES;
+		break;
+	case SCMP_FLTATR_ACT_BADARCH:
+		if (db_action_valid(value) == 0)
+			col->attr.act_badarch = value;
+		else
+			return -EINVAL;
+		break;
+	case SCMP_FLTATR_CTL_NNP:
+		col->attr.nnp_enable = (value ? 1 : 0);
+		break;
+	default:
+		return -EEXIST;
+		break;
+	}
+
+	return 0;
+}
+
+/**
+ * Add a new filter DB to a filter collection
+ * @param col the seccomp filter collection
+ * @param db the seccomp filter DB
+ *
+ * This function adds an existing seccomp filter DB to an existing seccomp
+ * filter collection assuming there isn't a filter DB already present with the
+ * same architecture.  Returns zero on success, negative values on failure.
+ *
+ */
+int db_col_db_add(struct db_filter_col *col, struct db_filter *db)
+{
+	struct db_filter **dbs;
+
+	if (db_col_arch_exist(col, db->arch->token))
+		return -EEXIST;
+
+	dbs = realloc(col->filters,
+		      sizeof(struct db_filter *) * (col->filter_cnt + 1));
+	if (dbs == NULL)
+		return -ENOMEM;
+	col->filters = dbs;
+	col->filter_cnt++;
+	col->filters[col->filter_cnt - 1] = db;
+
+	return 0;
+}
+
+/**
+ * Remove a filter DB from a filter collection
+ * @param col the seccomp filter collection
+ * @param arch_token the architecture token
+ *
+ * This function removes an existing seccomp filter DB from an existing seccomp
+ * filter collection.  Returns zero on success, negative values on failure.
+ *
+ */
+int db_col_db_remove(struct db_filter_col *col, uint32_t arch_token)
+{
+	unsigned int iter;
+	unsigned int found;
+	struct db_filter **dbs;
+
+	if ((col->filter_cnt <= 1) || (db_col_arch_exist(col, arch_token) == 0))
+		return -EINVAL;
+
+	for (found = 0, iter = 0; iter < col->filter_cnt; iter++) {
+		if (found)
+			col->filters[iter - 1] = col->filters[iter];
+		else if (col->filters[iter]->arch->token == arch_token) {
+			db_release(col->filters[iter]);
+			found = 1;
+		}
+	}
+	col->filters[--col->filter_cnt] = NULL;
+
+	/* NOTE: if we can't do the realloc it isn't fatal, we just have some
+	 *       extra space that will get cleaned up later */
+	dbs = realloc(col->filters,
+		      sizeof(struct db_filter *) * col->filter_cnt);
+	if (dbs != NULL)
+		col->filters = dbs;
+
+	return 0;
+}
+
 /**
  * Free and reset the seccomp filter DB
  * @param db the seccomp filter DB
@@ -319,7 +610,7 @@ int db_action_valid(uint32_t action)
  * default state; only the DB architecture is preserved.
  *
  */
-void db_reset(struct db_filter *db, uint32_t def_action)
+void db_reset(struct db_filter *db)
 {
 	struct db_sys_list *s_iter;
 
@@ -337,26 +628,17 @@ void db_reset(struct db_filter *db, uint32_t def_action)
 		}
 		db->syscalls = NULL;
 	}
-
-	/* set the default attribute values */
-	db->attr.act_default = def_action;
-	db->attr.act_badarch = SCMP_ACT_KILL;
-	db->attr.nnp_enable = 1;
-
-	/* set the state */
-	db->state = _DB_STA_VALID;
 }
 
 /**
  * Intitalize a seccomp filter DB
  * @param arch the architecture definition
- * @param def_action the default filter action
  *
  * This function initializes a seccomp filter DB and readies it for use.
  * Returns a pointer to the DB on success, NULL on failure.
  *
  */
-struct db_filter *db_init(const struct arch_def *arch, uint32_t def_action)
+struct db_filter *db_init(const struct arch_def *arch)
 {
 	struct db_filter *db;
 
@@ -369,7 +651,7 @@ struct db_filter *db_init(const struct arch_def *arch, uint32_t def_action)
 	db->arch = arch;
 
 	/* reset the DB to a known state */
-	db_reset(db, def_action);
+	db_reset(db);
 
 	return db;
 }
@@ -387,95 +669,11 @@ void db_release(struct db_filter *db)
 	if (db == NULL)
 		return;
 
-	/* set the state, just in case */
-	db->state = _DB_STA_FREED;
-
 	/* free and reset the DB */
-	db_reset(db, 0);
+	db_reset(db);
 	free(db);
 }
 
-/**
- * Validate a filter DB
- * @param db the seccomp filter DB
- *
- * This function validates a seccomp filter DB.  Returns zero if the DB is
- * valid, negative values on failure.
- *
- */
-int db_valid(struct db_filter *db)
-{
-	if (db != NULL && db->state == _DB_STA_VALID)
-		return 0;
-	return -EINVAL;
-}
-
-/**
- * Get a filter attribute
- * @param db the seccomp filter DB
- * @param attr the filter attribute
- * @param value the filter attribute value
- *
- * Get the requested filter attribute and provide it via @value.  Returns zero
- * on success, negative values on failure.
- *
- */
-int db_attr_get(const struct db_filter *db,
-		enum scmp_filter_attr attr, uint32_t *value)
-{
-	switch (attr) {
-	case SCMP_FLTATR_ACT_DEFAULT:
-		*value = db->attr.act_default;
-		break;
-	case SCMP_FLTATR_ACT_BADARCH:
-		*value = db->attr.act_badarch;
-		break;
-	case SCMP_FLTATR_CTL_NNP:
-		*value = db->attr.nnp_enable;
-		break;
-	default:
-		return -EEXIST;
-		break;
-	}
-
-	return 0;
-}
-
-/**
- * Set a filter attribute
- * @param db the seccomp filter DB
- * @param attr the filter attribute
- * @param value the filter attribute value
- *
- * Set the requested filter attribute with the given value.  Returns zero on
- * success, negative values on failure.
- *
- */
-int db_attr_set(struct db_filter *db,
-		enum scmp_filter_attr attr, uint32_t value)
-{
-	switch (attr) {
-	case SCMP_FLTATR_ACT_DEFAULT:
-		/* read only */
-		return -EACCES;
-		break;
-	case SCMP_FLTATR_ACT_BADARCH:
-		if (db_action_valid(value) == 0)
-			db->attr.act_badarch = value;
-		else
-			return -EINVAL;
-		break;
-	case SCMP_FLTATR_CTL_NNP:
-		db->attr.nnp_enable = (value ? 1 : 0);
-		break;
-	default:
-		return -EEXIST;
-		break;
-	}
-
-	return 0;
-}
-
 /**
  * Update the user specified portion of the syscall priority
  * @param db the seccomp filter db
@@ -570,7 +768,7 @@ static struct db_sys_list *_db_rule_gen_64(const struct arch_def *arch,
 	int chain_len_max;
 	struct db_sys_list *s_new;
 	struct db_arg_chain_tree *c_iter_hi = NULL, *c_iter_lo = NULL;
-	struct db_arg_chain_tree *c_prev_hi = NULL, *c_prev_lo = NULL;
+	struct db_arg_chain_tree *c_prev_lo = NULL;
 	unsigned int tf_flag;
 
 	s_new = malloc(sizeof(*s_new));
@@ -598,15 +796,12 @@ static struct db_sys_list *_db_rule_gen_64(const struct arch_def *arch,
 		memset(c_iter_lo, 0, sizeof(*c_iter_lo));
 		c_iter_lo->refcnt = 1;
 
-		/* link the hi node to the previous level */
-		if (c_prev_hi != NULL) {
+		/* link this level to the previous level */
+		if (c_prev_lo != NULL) {
 			if (tf_flag)
 				c_prev_lo->nxt_t = c_iter_hi;
-			else {
-				c_iter_hi->refcnt++;
-				c_prev_hi->nxt_f = c_iter_hi;
+			else
 				c_prev_lo->nxt_f = c_iter_hi;
-			}
 		} else
 			s_new->chains = c_iter_hi;
 		s_new->node_cnt += 2;
@@ -656,7 +851,6 @@ static struct db_sys_list *_db_rule_gen_64(const struct arch_def *arch,
 		/* link the hi and lo chain nodes */
 		c_iter_hi->nxt_t = c_iter_lo;
 
-		c_prev_hi = c_iter_hi;
 		c_prev_lo = c_iter_lo;
 	}
 	if (c_iter_lo != NULL) {
@@ -961,7 +1155,7 @@ add_reset:
 					ec_iter->act_f = action;
 				}
 				s_iter->node_cnt -= n_cnt;
-				return 0;
+				goto add_free_ok;
 			} else if (c_iter->nxt_t != NULL) {
 				if (ec_iter->nxt_t != NULL) {
 					/* jump to the next level */
@@ -976,6 +1170,7 @@ add_reset:
 					goto add_free_exist;
 				} else {
 					/* add a new branch */
+					c_prev = c_iter;
 					ec_iter->nxt_t = c_iter->nxt_t;
 					s_iter->node_cnt += (s_new->node_cnt-1);
 					goto add_free_match;
@@ -994,6 +1189,7 @@ add_reset:
 					goto add_free_exist;
 				} else {
 					/* add a new branch */
+					c_prev = c_iter;
 					ec_iter->nxt_f = c_iter->nxt_f;
 					s_iter->node_cnt += (s_new->node_cnt-1);
 					goto add_free_match;
