@@ -119,6 +119,8 @@ struct db_sys_list {
 	uint32_t action;
 
 	struct db_sys_list *next;
+	/* temporary use only by the BPF generator */
+	struct db_sys_list *pri_prv, *pri_nxt;
 };
 
 struct db_filter_attr {
@@ -131,18 +133,25 @@ struct db_filter_attr {
 };
 
 struct db_filter {
-	/* verification / state */
-	int state;
-
 	/* target architecture */
 	const struct arch_def *arch;
-	/* attributes */
-	struct db_filter_attr attr;
 
 	/* syscall filters, kept as a sorted single-linked list */
 	struct db_sys_list *syscalls;
 };
 
+struct db_filter_col {
+	/* verification / state */
+	int state;
+
+	/* attributes */
+	struct db_filter_attr attr;
+
+	/* individual filters */
+	struct db_filter **filters;
+	unsigned int filter_cnt;
+};
+
 /**
  * Iterate over each item in the DB list
  * @param iter the iterator
@@ -157,17 +166,28 @@ struct db_filter {
 
 int db_action_valid(uint32_t action);
 
-void db_reset(struct db_filter *db, uint32_t def_action);
-struct db_filter *db_init(const struct arch_def *arch, uint32_t def_action);
+struct db_filter_col *db_col_init(uint32_t def_action);
+void db_col_reset(struct db_filter_col *col, uint32_t def_action);
+void db_col_release(struct db_filter_col *col);
+
+int db_col_valid(struct db_filter_col *col);
+
+int db_col_merge(struct db_filter_col *col_dst, struct db_filter_col *col_src);
+
+int db_col_arch_exist(struct db_filter_col *col, uint32_t arch_token);
+
+int db_col_attr_get(const struct db_filter_col *col,
+		    enum scmp_filter_attr attr, uint32_t *value);
+int db_col_attr_set(struct db_filter_col *col,
+		    enum scmp_filter_attr attr, uint32_t value);
+
+int db_col_db_add(struct db_filter_col *col, struct db_filter *db);
+int db_col_db_remove(struct db_filter_col *col, uint32_t arch_token);
+
+struct db_filter *db_init(const struct arch_def *arch);
+void db_reset(struct db_filter *db);
 void db_release(struct db_filter *db);
 
-int db_valid(struct db_filter *db);
-
-int db_attr_get(const struct db_filter *db,
-		enum scmp_filter_attr attr, uint32_t *value);
-int db_attr_set(struct db_filter *db,
-		enum scmp_filter_attr attr, uint32_t value);
-
 int db_syscall_priority(struct db_filter *db,
 			unsigned int syscall, uint8_t priority);
 
