@@ -58,19 +58,21 @@ struct bpf_jump {
 		uint8_t imm_j;
 		uint32_t imm_k;
 		uint64_t hash;
-		void *ptr;
+		struct db_arg_chain_tree *db;
+		struct bpf_blk *blk;
+		unsigned int nxt;
 	} tgt;
 };
 #define _BPF_JMP_NO \
-	((struct bpf_jump) { TGT_NONE, { .ptr = 0 } })
-#define _BPF_JMP_NXT \
-	((struct bpf_jump) { TGT_NXT, { .ptr = 0 } })  /* be careful! */
+	((struct bpf_jump) { TGT_NONE })
+#define _BPF_JMP_NXT(x) \
+	((struct bpf_jump) { TGT_NXT, { .nxt = (x) } })  /* be careful! */
 #define _BPF_JMP_IMM(x) \
 	((struct bpf_jump) { TGT_IMM, { .imm_j = (x) } })
 #define _BPF_JMP_DB(x) \
-	((struct bpf_jump) { TGT_PTR_DB, { .ptr = (x) } })
+	((struct bpf_jump) { TGT_PTR_DB, { .db = (x) } })
 #define _BPF_JMP_BLK(x) \
-	((struct bpf_jump) { TGT_PTR_BLK, { .ptr = (x) } })
+	((struct bpf_jump) { TGT_PTR_BLK, { .blk = (x) } })
 #define _BPF_JMP_HSH(x) \
 	((struct bpf_jump) { TGT_PTR_HSH, { .hash = (x) } })
 #define _BPF_K(x) \
@@ -98,11 +100,12 @@ struct bpf_blk {
 	/* original db_arg_chain_tree node */
 	const struct db_arg_chain_tree *node;
 
-	/* used during final block assembly */
+	/* used during block assembly */
 	unsigned int hashed_flag;
 	uint64_t hash;
 	struct bpf_blk *hash_nxt;
 	struct bpf_blk *prev, *next;
+	struct bpf_blk *lvl_prv, *lvl_nxt;
 };
 #define _BLK_MSZE(x) \
 	((x)->blk_cnt * sizeof(*((x)->blks)))
@@ -118,14 +121,14 @@ struct bpf_hash_bkt {
 #define _BPF_HASH_SIZE			(1 << _BPF_HASH_BITS)
 #define _BPF_HASH_MASK			(_BPF_HASH_BITS - 1)
 struct bpf_state {
-	/* target arch */
-	const struct arch_def *arch;
 	/* filter attributes */
 	const struct db_filter_attr *attr;
-
 	/* default action */
 	uint64_t def_hsh;
 
+	/* target arch - NOTE: be careful, temporary use only! */
+	const struct arch_def *arch;
+
 	/* block hash table */
 	struct bpf_hash_bkt *htbl[_BPF_HASH_SIZE];
 
@@ -156,7 +159,7 @@ struct bpf_state {
 static struct bpf_blk *_gen_bpf_chain(struct bpf_state *state,
 				      const struct db_sys_list *sys,
 				      const struct db_arg_chain_tree *chain,
-				      struct bpf_blk *next);
+				      const struct bpf_jump *nxt_jump);
 
 static struct bpf_blk *_hsh_remove(struct bpf_state *state, uint64_t h_val);
 static struct bpf_blk *_hsh_find(const struct bpf_state *state, uint64_t h_val);
@@ -211,7 +214,7 @@ static void _blk_free(struct bpf_state *state, struct bpf_blk *blk)
 		i_iter = &blk->blks[iter];
 		switch (i_iter->jt.type) {
 		case TGT_PTR_BLK:
-			_blk_free(state, i_iter->jt.tgt.ptr);
+			_blk_free(state, i_iter->jt.tgt.blk);
 			break;
 		case TGT_PTR_HSH:
 			b_iter = _hsh_find(state, i_iter->jt.tgt.hash);
@@ -223,7 +226,7 @@ static void _blk_free(struct bpf_state *state, struct bpf_blk *blk)
 		}
 		switch (i_iter->jf.type) {
 		case TGT_PTR_BLK:
-			_blk_free(state, i_iter->jf.tgt.ptr);
+			_blk_free(state, i_iter->jf.tgt.blk);
 			break;
 		case TGT_PTR_HSH:
 			b_iter = _hsh_find(state, i_iter->jf.tgt.hash);
@@ -245,7 +248,7 @@ static void _blk_free(struct bpf_state *state, struct bpf_blk *blk)
  *
  * Add the new BPF instruction to the end of the give instruction block.  If
  * the given instruction block is NULL, a new block will be allocated.  Returns
- * a pointer to the block on success, NULL on failure and in the case of
+ * a pointer to the block on success, NULL on failure, and in the case of
  * failure the instruction block is free'd.
  *
  */
@@ -275,6 +278,51 @@ static struct bpf_blk *_blk_append(struct bpf_state *state,
 	return blk;
 }
 
+#if 0 /* see NOTE in function description */
+/**
+ * Append a BPF instruction block to another BPF instruction block
+ * @param state the BPF state
+ * @param dst the destination instruction block
+ * @param src the source instruction block
+ *
+ * Append the source instruction block to the end of the destination block.
+ * Returns a pointer to dst on success, NULL on failure, and in the case of
+ * failure the destination block is free'd.
+ *
+ * NOTE: We're going to #if'def this function out and not simply remove it
+ *	 because the need for such a function does seem to reoccur from time
+ *	 to time and I don't feel like re-implementing it again in the near
+ *	 future.  Once we've given the BPF generator a good scrubbing we can
+ *	 remove this function if it still isn't used.
+ *
+ */
+static struct bpf_blk *_blk_append_blk(struct bpf_state *state,
+				       struct bpf_blk *dst,
+				       const struct bpf_blk *src)
+{
+	unsigned int size;
+	struct bpf_instr *new;
+
+	size = dst->blk_cnt + src->blk_cnt;
+	if (size > dst->blk_alloc) {
+		dst->blk_alloc = size;
+		new = realloc(dst->blks, dst->blk_alloc * sizeof(*(dst->blks)));
+		if (new == NULL) {
+			_blk_free(state, dst);
+			return NULL;
+		}
+		dst->blks = new;
+	}
+	memcpy(&dst->blks[dst->blk_cnt], src->blks,
+	       src->blk_cnt * sizeof(*(src->blks)));
+	dst->blk_cnt += src->blk_cnt;
+	if (src->priority > dst->priority)
+		dst->priority = src->priority;
+
+	return dst;
+}
+#endif /* see NOTE in function description */
+
 /**
  * Append a block of BPF instructions to the final BPF program
  * @param prg the BPF program
@@ -355,29 +403,6 @@ bpf_append_blk_failure:
 	return rc;
 }
 
-/**
- * Append a single BPF instruction to the final BPF program
- * @param prg the BPF program
- * @param instr the BPF instruction
- *
- * Add the BPF instruction to the end of the BPF program and perform the
- * necssary translation.  Returns zero on success, negative values on failure
- * and in the case of failure the BPF program is free'd.
- *
- */
-static int _bpf_append_instr(struct bpf_program *prg,
-			     struct bpf_instr *instr)
-{
-	struct bpf_blk blk;
-
-	memset(&blk, 0, sizeof(blk));
-	blk.blk_cnt = 1;
-	blk.blk_alloc = 1;
-	blk.blks = instr;
-
-	return _bpf_append_blk(prg, &blk);
-}
-
 /**
  * Free the BPF program
  * @param prg the BPF program
@@ -748,13 +773,13 @@ static struct bpf_blk *_gen_bpf_node(struct bpf_state *state,
 	else if (node->act_t_flg)
 		instr.jt = _BPF_JMP_HSH(act_t_hash);
 	else
-		instr.jt = _BPF_JMP_NXT;
+		instr.jt = _BPF_JMP_NXT(0);
 	if (node->nxt_f != NULL)
 		instr.jf = _BPF_JMP_DB(node->nxt_f);
 	else if (node->act_f_flg)
 		instr.jf = _BPF_JMP_HSH(act_f_hash);
 	else
-		instr.jf = _BPF_JMP_NXT;
+		instr.jf = _BPF_JMP_NXT(0);
 	blk = _blk_append(state, blk, &instr);
 	if (blk == NULL)
 		goto node_failure;
@@ -767,95 +792,12 @@ node_failure:
 	return NULL;
 }
 
-/**
- * Generate a BPF instruction block for a given filter DB level
- * @param state the BPF state
- * @param sys the syscall filter DB node
- * @param node the filter DB node
- * @param next the filter node to fallthrough to at the end of the level
- *
- * Generate a BPF instruction block which executes the filter specified by the
- * given filter DB level.  Returns a pointer to the instruction block on
- * success, NULL on failure.  The given BPF block is free'd on failure.
- *
- */
-static struct bpf_blk *_gen_bpf_chain_lvl(struct bpf_state *state,
-					  const struct db_sys_list *sys,
-					  const struct db_arg_chain_tree *node,
-					  struct bpf_blk *next)
-{
-	struct bpf_blk *blk;
-	struct bpf_blk *b_head = NULL, *b_prev = NULL, *b_next, *b_iter;
-	struct bpf_instr *i_iter;
-	const struct db_arg_chain_tree *l_iter;
-	struct acc_state a_state = { -1, ARG_MASK_MAX };
-	struct bpf_jump def_jump;
-	unsigned int iter;
-
-	if (node == NULL) {
-		blk = _gen_bpf_action(state, NULL, sys->action);
-		if (blk == NULL)
-			goto chain_lvl_failure;
-		return blk;
-	}
-
-	/* find the starting node of the level */
-	l_iter = node;
-	while (l_iter->lvl_prv != NULL)
-		l_iter = l_iter->lvl_prv;
-
-	/* build all of the blocks for this level */
-	do {
-		blk = _gen_bpf_node(state, l_iter, &a_state);
-		if (blk == NULL)
-			goto chain_lvl_failure;
-		if (b_head != NULL) {
-			b_prev->next = blk;
-			blk->prev = b_prev;
-		} else
-			b_head = blk;
-
-		b_prev = blk;
-		l_iter = l_iter->lvl_nxt;
-	} while (l_iter != NULL);
-
-	if (next != NULL)
-		def_jump = _BPF_JMP_BLK(next);
-	else
-		def_jump = _BPF_JMP_HSH(state->def_hsh);
-
-	/* resolve the TGT_NXT jumps */
-	b_iter = b_head;
-	do {
-		b_next = b_iter->next;
-		for (iter = 0; iter < b_iter->blk_cnt; iter++) {
-			i_iter = &b_iter->blks[iter];
-			if (i_iter->jt.type == TGT_NXT)
-				i_iter->jt = (b_next == NULL ?
-					      def_jump : _BPF_JMP_BLK(b_next));
-			if (i_iter->jf.type == TGT_NXT)
-				i_iter->jf = (b_next == NULL ?
-					      def_jump : _BPF_JMP_BLK(b_next));
-		}
-		b_iter = b_next;
-	} while (b_iter != NULL);
-
-	return b_head;
-
-chain_lvl_failure:
-	while (b_head != NULL) {
-		b_iter = b_head;
-		b_head = b_iter->next;
-		_blk_free(state, b_iter);
-	}
-	return NULL;
-}
-
 /**
  * Resolve the jump targets in a BPF instruction block
  * @param state the BPF state
  * @param sys the syscall filter
  * @param blk the BPF instruction block
+ * @param nxt_jump the jump to fallthrough to at the end of the level
  *
  * Resolve the jump targets in a BPF instruction block generated by the
  * _gen_bpf_chain_lvl() function and adds the resulting block to the hash
@@ -865,25 +807,18 @@ chain_lvl_failure:
  */
 static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
 					      const struct db_sys_list *sys,
-					      struct bpf_blk *blk)
+					      struct bpf_blk *blk,
+					      const struct bpf_jump *nxt_jump)
 {
 	int rc;
 	unsigned int iter;
 	struct bpf_blk *b_new;
 	struct bpf_instr *i_iter;
 	struct db_arg_chain_tree *node;
-	struct bpf_blk *next;
 
 	if (blk->hashed_flag)
 		return blk;
 
-	if (state->arch->size == ARCH_SIZE_64 && blk->node != NULL &&
-	    blk->node->arg_offset == arch_arg_offset_hi(state->arch,
-							blk->node->arg))
-		next = blk->next;
-	else
-		next = NULL;
-
 	/* convert TGT_PTR_DB to TGT_PTR_HSH references */
 	for (iter = 0; iter < blk->blk_cnt; iter++) {
 		i_iter = &blk->blks[iter];
@@ -895,14 +830,15 @@ static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
 			break;
 		case TGT_PTR_BLK:
 			b_new = _gen_bpf_chain_lvl_res(state, sys,
-						       i_iter->jt.tgt.ptr);
+						       i_iter->jt.tgt.blk,
+						       nxt_jump);
 			if (b_new == NULL)
 				return NULL;
 			i_iter->jt = _BPF_JMP_HSH(b_new->hash);
 			break;
 		case TGT_PTR_DB:
-			node = (struct db_arg_chain_tree *)i_iter->jt.tgt.ptr;
-			b_new = _gen_bpf_chain(state, sys, node, next);
+			node = (struct db_arg_chain_tree *)i_iter->jt.tgt.db;
+			b_new = _gen_bpf_chain(state, sys, node, nxt_jump);
 			if (b_new == NULL)
 				return NULL;
 			i_iter->jt = _BPF_JMP_HSH(b_new->hash);
@@ -919,14 +855,15 @@ static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
 			break;
 		case TGT_PTR_BLK:
 			b_new = _gen_bpf_chain_lvl_res(state, sys,
-						       i_iter->jf.tgt.ptr);
+						       i_iter->jf.tgt.blk,
+						       nxt_jump);
 			if (b_new == NULL)
 				return NULL;
 			i_iter->jf = _BPF_JMP_HSH(b_new->hash);
 			break;
 		case TGT_PTR_DB:
-			node = (struct db_arg_chain_tree *)i_iter->jf.tgt.ptr;
-			b_new = _gen_bpf_chain(state, sys, node, next);
+			node = (struct db_arg_chain_tree *)i_iter->jf.tgt.db;
+			b_new = _gen_bpf_chain(state, sys, node, nxt_jump);
 			if (b_new == NULL)
 				return NULL;
 			i_iter->jf = _BPF_JMP_HSH(b_new->hash);
@@ -941,13 +878,6 @@ static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
 		case TGT_PTR_HSH:
 			/* ignore these jump types */
 			break;
-		case TGT_PTR_DB:
-			node = (struct db_arg_chain_tree *)i_iter->k.tgt.ptr;
-			b_new = _gen_bpf_chain(state, sys, node, NULL);
-			if (b_new == NULL)
-				return NULL;
-			i_iter->k = _BPF_JMP_HSH(b_new->hash);
-			break;
 		default:
 			/* we should not be here */
 			return NULL;
@@ -967,7 +897,7 @@ static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
  * @param state the BPF state
  * @param sys the syscall filter
  * @param chain the filter chain
- * @param next the filter node to fallthrough to at the end of the level
+ * @param nxt_jump the jump to fallthrough to at the end of the level
  *
  * Generate the BPF instruction blocks for the given filter chain and return
  * a pointer to the first block on success; returns NULL on failure.
@@ -976,51 +906,306 @@ static struct bpf_blk *_gen_bpf_chain_lvl_res(struct bpf_state *state,
 static struct bpf_blk *_gen_bpf_chain(struct bpf_state *state,
 				      const struct db_sys_list *sys,
 				      const struct db_arg_chain_tree *chain,
-				      struct bpf_blk *next)
+				      const struct bpf_jump *nxt_jump)
 {
-	struct bpf_blk *blk;
+	struct bpf_blk *b_head = NULL, *b_tail = NULL;
+	struct bpf_blk *b_prev, *b_next, *b_iter;
+	struct bpf_instr *i_iter;
+	const struct db_arg_chain_tree *c_iter;
+	struct acc_state a_state = { -1, ARG_MASK_MAX };
+	unsigned int iter;
 
-	blk = _gen_bpf_chain_lvl(state, sys, chain, next);
-	if (blk == NULL)
-		return NULL;
-	return _gen_bpf_chain_lvl_res(state, sys, blk);
+	if (chain == NULL) {
+		b_head = _gen_bpf_action(state, NULL, sys->action);
+		if (b_head == NULL)
+			goto chain_failure;
+		b_tail = b_head;
+	} else {
+		/* find the starting node of the level */
+		c_iter = chain;
+		while (c_iter->lvl_prv != NULL)
+			c_iter = c_iter->lvl_prv;
+
+		/* build all of the blocks for this level */
+		do {
+			b_iter = _gen_bpf_node(state, c_iter, &a_state);
+			if (b_iter == NULL)
+				goto chain_failure;
+			if (b_head != NULL) {
+				b_iter->lvl_prv = b_tail;
+				b_tail->lvl_nxt = b_iter;
+				b_tail = b_iter;
+			} else {
+				b_head = b_iter;
+				b_tail = b_iter;
+			}
+			c_iter = c_iter->lvl_nxt;
+		} while (c_iter != NULL);
+
+		/* resolve the TGT_NXT jumps */
+		b_iter = b_head;
+		do {
+			b_next = b_iter->lvl_nxt;
+			for (iter = 0; iter < b_iter->blk_cnt; iter++) {
+				i_iter = &b_iter->blks[iter];
+				if (i_iter->jt.type == TGT_NXT) {
+					if (i_iter->jt.tgt.nxt != 0)
+						goto chain_failure;
+					i_iter->jt = (b_next == NULL ?
+						      *nxt_jump :
+						      _BPF_JMP_BLK(b_next));
+				}
+				if (i_iter->jf.type == TGT_NXT) {
+					if (i_iter->jf.tgt.nxt != 0)
+						goto chain_failure;
+					i_iter->jf = (b_next == NULL ?
+						      *nxt_jump :
+						      _BPF_JMP_BLK(b_next));
+				}
+			}
+			b_iter = b_next;
+		} while (b_iter != NULL);
+	}
+
+	/* resolve all of the blocks */
+	b_iter = b_tail;
+	do {
+		/* b_iter may change after resolving, so save the linkage */
+		b_prev = b_iter->lvl_prv;
+		b_next = b_iter->lvl_nxt;
+
+		b_iter = _gen_bpf_chain_lvl_res(state, sys, b_iter,
+						(b_next == NULL ?
+						 nxt_jump :
+						 &_BPF_JMP_BLK(b_next)));
+		if (b_iter == NULL)
+			goto chain_failure;
+
+		/* restore the block linkage on this level */
+		if (b_prev != NULL)
+			b_prev->lvl_nxt = b_iter;
+		b_iter->lvl_prv = b_prev;
+		b_iter->lvl_nxt = b_next;
+		if (b_next != NULL)
+			b_next->lvl_prv = b_iter;
+		if (b_iter->lvl_prv == NULL)
+			b_head = b_iter;
+
+		b_iter = b_prev;
+	} while (b_iter != NULL);
+
+	return b_head;
+
+chain_failure:
+	while (b_head != NULL) {
+		b_iter = b_head;
+		b_head = b_iter->lvl_nxt;
+		_blk_free(state, b_iter);
+	}
+	return NULL;
 }
 
 /**
  * Generate the BPF instruction blocks for a given syscall
  * @param state the BPF state
  * @param sys the syscall filter DB entry
+ * @param nxt_hash the hash value of the next syscall filter DB entry
+ * @param first this is the first syscall for the architecture
  *
  * Generate the BPF instruction blocks for the given syscall filter and return
- * a pointer to the first block on success; returns NULL on failure.
+ * a pointer to the first block on success; returns NULL on failure.  It is
+ * important to note that the block returned has not been added to the hash
+ * table, however, any linked/referenced blocks have been added to the hash
+ * table.
  *
  */
 static struct bpf_blk *_gen_bpf_syscall(struct bpf_state *state,
-					const struct db_sys_list *sys)
+					const struct db_sys_list *sys,
+					uint64_t nxt_hash, int first)
 {
 	int rc;
 	struct bpf_instr instr;
-	struct bpf_blk *blk_c, *blk_s;
+	struct bpf_blk *blk_c, *blk_s = NULL;
+	struct bpf_jump def_jump = _BPF_JMP_HSH(state->def_hsh);
 
 	/* generate the argument chains */
-	blk_c = _gen_bpf_chain(state, sys, sys->chains, NULL);
+	blk_c = _gen_bpf_chain(state, sys, sys->chains, &def_jump);
 	if (blk_c == NULL)
 		return NULL;
 
-	/* syscall check (syscall number is still in the accumulator) */
+	/* load the syscall into the accumulator (if needed) */
+	if (first) {
+		_BPF_INSTR(instr, BPF_LD+BPF_ABS, _BPF_JMP_NO, _BPF_JMP_NO,
+			_BPF_SYSCALL);
+		blk_s = _blk_append(state, NULL, &instr);
+		if (blk_s == NULL)
+			return NULL;
+	}
+
+	/* syscall check */
 	_BPF_INSTR(instr, BPF_JMP+BPF_JEQ,
-		   _BPF_JMP_HSH(blk_c->hash), _BPF_JMP_NXT, _BPF_K(sys->num));
-	blk_s = _blk_append(state, NULL, &instr);
+		   _BPF_JMP_HSH(blk_c->hash), _BPF_JMP_HSH(nxt_hash),
+		   _BPF_K(sys->num));
+	blk_s = _blk_append(state, blk_s, &instr);
 	if (blk_s == NULL)
 		return NULL;
 	blk_s->priority = sys->priority;
+
+	/* add to the hash table */
 	rc = _hsh_add(state, &blk_s, 1);
-	if (rc < 0)
+	if (rc < 0) {
+		_blk_free(state, blk_s);
 		return NULL;
+	}
 
 	return blk_s;
 }
 
+/**
+ * Generate the BPF instruction blocks for a given filter/architecture
+ * @param state the BPF state
+ * @param db the filter DB
+ *
+ * Generate the BPF instruction block for the given filter DB/architecture and
+ * return a pointer to the block on succes, NULL on failure.  The resulting
+ * block assumes that the architecture token has already been loaded into the
+ * BPF accumulator.
+ *
+ */
+static struct bpf_blk *_gen_bpf_arch(struct bpf_state *state,
+				     const struct db_filter *db)
+{
+	int rc;
+	unsigned int blk_cnt = 0;
+	struct bpf_instr instr;
+	struct db_sys_list *s_head = NULL, *s_tail = NULL, *s_iter, *s_iter_b;
+	struct bpf_blk *b_head = NULL, *b_tail = NULL, *b_iter, *b_new;
+
+	state->arch = db->arch;
+
+	/* sort the syscall list */
+	db_list_foreach(s_iter, db->syscalls) {
+		if (s_head != NULL) {
+			s_iter_b = s_head;
+			while ((s_iter_b->pri_nxt != NULL) &&
+			       (s_iter->priority <= s_iter_b->priority))
+				s_iter_b = s_iter_b->pri_nxt;
+
+			if (s_iter->priority > s_iter_b->priority) {
+				s_iter->pri_prv = s_iter_b->pri_prv;
+				s_iter->pri_nxt = s_iter_b;
+				if (s_iter_b == s_head) {
+					s_head->pri_prv = s_iter;
+					s_head = s_iter;
+				} else {
+					s_iter->pri_prv->pri_nxt = s_iter;
+					s_iter->pri_nxt->pri_prv = s_iter;
+				}
+			} else {
+				s_iter->pri_prv = s_tail;
+				s_iter->pri_nxt = NULL;
+				s_iter->pri_prv->pri_nxt = s_iter;
+				s_tail = s_iter;
+			}
+		} else {
+			s_head = s_iter;
+			s_tail = s_iter;
+			s_head->pri_prv = NULL;
+			s_head->pri_nxt = NULL;
+		}
+	}
+
+	/* create the syscall filters and add them to block list group */
+	for (s_iter = s_tail; s_iter != NULL; s_iter = s_iter->pri_prv) {
+		if (s_iter->valid == 0)
+			continue;
+
+		/* build the syscall filter */
+		b_new = _gen_bpf_syscall(state, s_iter,
+					 (b_head == NULL ?
+					  state->def_hsh : b_head->hash),
+					 (s_iter == s_head ? 1 : 0));
+		if (b_new == NULL)
+			goto arch_failure;
+
+		/* add the filter to the list head */
+		b_new->prev = NULL;
+		b_new->next = b_head;
+		if (b_tail != NULL) {
+			b_head->prev = b_new;
+			b_head = b_new;
+		} else {
+			b_head = b_new;
+			b_tail = b_head;
+		}
+
+		if (b_tail->next != NULL)
+			b_tail = b_tail->next;
+		blk_cnt++;
+	}
+
+	/* do the architecture check and load the syscall number */
+	if (b_head != NULL) {
+		/* arch check */
+		_BPF_INSTR(instr, BPF_JMP+BPF_JEQ,
+			   _BPF_JMP_HSH(b_head->hash),
+			   _BPF_JMP_NXT(blk_cnt),
+			   _BPF_K(db->arch->token));
+		b_head->prev = _blk_append(state, NULL, &instr);
+		if (b_head->prev == NULL)
+			goto arch_failure;
+		b_head->prev->next = b_head;
+		b_head = b_head->prev;
+	} else {
+		/* arch check */
+		_BPF_INSTR(instr, BPF_JMP+BPF_JEQ,
+			   _BPF_JMP_HSH(state->def_hsh), _BPF_JMP_NXT(0),
+			   _BPF_K(db->arch->token));
+		b_head = _blk_append(state, NULL, &instr);
+		if (b_head == NULL)
+			goto arch_failure;
+	}
+	rc = _hsh_add(state, &b_head, 1);
+	if (rc < 0)
+		goto arch_failure;
+
+	state->arch = NULL;
+	return b_head;
+
+arch_failure:
+	/* NOTE: we do the cleanup here and not just return an error as all of
+	 * the instruction blocks may not be added to the hash table when we
+	 * hit an error */
+	state->arch = NULL;
+	b_iter = b_head;
+	while (b_iter != NULL) {
+		b_new = b_iter->next;
+		_blk_free(state, b_iter);
+		b_iter = b_new;
+	}
+	return NULL;
+}
+
+/**
+ * Find the target block for the "next" jump
+ * @param blk the instruction block
+ * @param nxt the next offset
+ *
+ * Find the target block for the TGT_NXT jump using the given offset.  Returns
+ * a pointer to the target block on success or NULL on failure.
+ *
+ */
+static struct bpf_blk *_gen_bpf_find_nxt(const struct bpf_blk *blk,
+					 unsigned int nxt)
+{
+	struct bpf_blk *iter = blk->next;
+
+	for (; (iter != NULL) && (nxt > 0); nxt--)
+		iter = iter->next;
+
+	return iter;
+}
+
 /**
  * Manage jumps to return instructions
  * @param state the BPF state
@@ -1178,16 +1363,16 @@ static int _gen_bpf_build_jmp(struct bpf_state *state,
 }
 
 /**
- * Generate the BPF program for the given filter DB
+ * Generate the BPF program for the given filter collection
  * @param state the BPF state
- * @param db the filter DB
+ * @param col the filter collection
  *
- * Generate the BPF program for the given filter DB.  Returns zero on success,
- * negative values on failure.
+ * Generate the BPF program for the given filter collection.  Returns zero on
+ * success, negative values on failure.
  *
  */
 static int _gen_bpf_build_bpf(struct bpf_state *state,
-			      const struct db_filter *db)
+			      const struct db_filter_col *col)
 {
 	int rc;
 	int iter;
@@ -1196,83 +1381,77 @@ static int _gen_bpf_build_bpf(struct bpf_state *state,
 	unsigned int jmp_len;
 	struct bpf_instr instr;
 	struct bpf_instr *i_iter;
-	struct db_sys_list *s_iter;
-	struct bpf_blk *def_blk;
+	struct bpf_blk *b_badarch, *b_default;
 	struct bpf_blk *b_head = NULL, *b_tail = NULL, *b_iter, *b_new, *b_jmp;
 
-	/* create the default action */
-	def_blk = _gen_bpf_action(state, NULL, state->attr->act_default);
-	if (def_blk == NULL)
+	if (col->filter_cnt == 0)
+		return -EINVAL;
+
+	/* generate the badarch action */
+	b_badarch = _gen_bpf_action(state, NULL, state->attr->act_badarch);
+	if (b_badarch == NULL)
 		return -ENOMEM;
-	rc = _hsh_add(state, &def_blk, 1);
+	rc = _hsh_add(state, &b_badarch, 1);
 	if (rc < 0)
 		return rc;
-	state->def_hsh = def_blk->hash;
 
-	/* create the syscall filters and add them to the top block group */
-	db_list_foreach(s_iter, db->syscalls) {
-		if (s_iter->valid == 0)
-			continue;
+	/* generate the default action */
+	b_default = _gen_bpf_action(state, NULL, state->attr->act_default);
+	if (b_default == NULL)
+		return -ENOMEM;
+	rc = _hsh_add(state, &b_default, 0);
+	if (rc < 0)
+		return rc;
+	state->def_hsh = b_default->hash;
 
-		/* build the syscall filter */
-		b_new = _gen_bpf_syscall(state, s_iter);
+	/* load the architecture token/number */
+	_BPF_INSTR(instr, BPF_LD+BPF_ABS, _BPF_JMP_NO, _BPF_JMP_NO,
+		   _BPF_K(offsetof(struct seccomp_data, arch)));
+	b_head = _blk_append(state, NULL, &instr);
+	if (b_head == NULL)
+		return -ENOMEM;
+	rc = _hsh_add(state, &b_head, 1);
+	if (rc < 0)
+		return rc;
+	b_tail = b_head;
+
+	/* generate the per-architecture filters */
+	for (iter = 0; iter < col->filter_cnt; iter++) {
+		b_new = _gen_bpf_arch(state, col->filters[iter]);
 		if (b_new == NULL)
 			return -ENOMEM;
-		/* add the filter to the list, sorting based on priority */
-		if (b_head != NULL) {
-			b_iter = b_head;
-			do {
-				if (b_new->priority > b_iter->priority) {
-					if (b_iter == b_head) {
-						b_new->next = b_head;
-						b_head->prev = b_new;
-						b_head = b_new;
-					} else {
-						b_iter->prev->next = b_new;
-						b_new->prev = b_iter->prev;
-						b_new->next = b_iter;
-						b_iter->prev = b_new;
-					}
-					b_iter = NULL;
-				} else {
-					if (b_iter->next == NULL) {
-						b_iter->next = b_new;
-						b_new->prev = b_iter;
-						b_iter = NULL;
-					} else
-						b_iter = b_iter->next;
-				}
-			} while (b_iter != NULL);
-			if (b_tail->next != NULL)
-				b_tail = b_tail->next;
-		} else {
-			b_head = b_new;
-			b_tail = b_head;
-			b_head->prev = NULL;
-			b_head->next = NULL;
-		}
+		b_new->prev = b_tail;
+		b_tail->next = b_new;
+		b_tail = b_new;
+		while (b_tail->next != NULL)
+			b_tail = b_tail->next;
 	}
 
-	/* tack on the default action to the end of the top block group */
-	if (b_tail != NULL) {
-		b_tail->next = def_blk;
-		def_blk->prev = b_tail;
-		b_tail = def_blk;
-	} else {
-		b_head = def_blk;
-		b_tail = b_head;
-	}
+	/* add a badarch action to the end */
+	b_badarch->prev = b_tail;
+	b_badarch->next = NULL;
+	b_tail->next = b_badarch;
+	b_tail = b_badarch;
 
 	/* resolve any TGT_NXT jumps at the top level */
 	b_iter = b_head;
 	do {
-		b_jmp = b_iter->next;
 		for (iter = 0; iter < b_iter->blk_cnt; iter++) {
 			i_iter = &b_iter->blks[iter];
-			if (i_iter->jt.type == TGT_NXT)
+			if (i_iter->jt.type == TGT_NXT) {
+				b_jmp = _gen_bpf_find_nxt(b_iter,
+							  i_iter->jt.tgt.nxt);
+				if (b_jmp == NULL)
+					return -EFAULT;
 				i_iter->jt = _BPF_JMP_HSH(b_jmp->hash);
-			if (i_iter->jf.type == TGT_NXT)
+			}
+			if (i_iter->jf.type == TGT_NXT) {
+				b_jmp = _gen_bpf_find_nxt(b_iter,
+							  i_iter->jf.tgt.nxt);
+				if (b_jmp == NULL)
+					return -EFAULT;
 				i_iter->jf = _BPF_JMP_HSH(b_jmp->hash);
+			}
 			/* we shouldn't need to worry about a TGT_NXT in k */
 		}
 		b_iter = b_iter->next;
@@ -1283,7 +1462,9 @@ static int _gen_bpf_build_bpf(struct bpf_state *state,
 	do {
 		b_jmp = NULL;
 		/* look for jumps - backwards (shorter jumps) */
-		for (iter = b_iter->blk_cnt - 1; iter >= 0; iter--) {
+		for (iter = b_iter->blk_cnt - 1;
+		     (iter >= 0) && (b_jmp == NULL);
+		     iter--) {
 			i_iter = &b_iter->blks[iter];
 			if (i_iter->jt.type == TGT_PTR_HSH)
 				b_jmp = _hsh_find_once(state,
@@ -1301,7 +1482,6 @@ static int _gen_bpf_build_bpf(struct bpf_state *state,
 				b_iter->next = b_jmp;
 				if (b_jmp->next)
 					b_jmp->next->prev=b_jmp;
-				break;
 			}
 		}
 		if (b_jmp != NULL) {
@@ -1312,32 +1492,6 @@ static int _gen_bpf_build_bpf(struct bpf_state *state,
 			b_iter = b_iter->prev;
 	} while (b_iter != NULL);
 
-	/* verify the architecture and kill ourselves if different */
-	_BPF_INSTR(instr, BPF_LD+BPF_ABS,
-		   _BPF_JMP_NO, _BPF_JMP_NO,
-		   _BPF_K(offsetof(struct seccomp_data, arch)));
-	rc = _bpf_append_instr(state->bpf, &instr);
-	if (rc < 0)
-		return rc;
-	_BPF_INSTR(instr, BPF_JMP+BPF_JEQ,
-		   _BPF_JMP_IMM(1), _BPF_JMP_NO,
-		   _BPF_K(state->arch->token));
-	rc = _bpf_append_instr(state->bpf, &instr);
-	if (rc < 0)
-		return rc;
-	_BPF_INSTR(instr, BPF_RET,
-		   _BPF_JMP_NO, _BPF_JMP_NO, _BPF_K(state->attr->act_badarch));
-	rc = _bpf_append_instr(state->bpf, &instr);
-	if (rc < 0)
-		return rc;
-
-	/* load the syscall into the accumulator */
-	_BPF_INSTR(instr, BPF_LD+BPF_ABS,
-		   _BPF_JMP_NO, _BPF_JMP_NO, _BPF_SYSCALL);
-	rc = _bpf_append_instr(state->bpf, &instr);
-	if (rc < 0)
-		return rc;
-
 	/* NOTE - from here to the end of the function we need to fail via the
 	 *	  the build_bpf_free_blks label, not just return an error; see
 	 *	  the _gen_bpf_build_jmp() function for details */
@@ -1459,27 +1613,26 @@ build_bpf_free_blks:
 
 /**
  * Generate a BPF representation of the filter DB
- * @param db the seccomp filter DB
+ * @param col the seccomp filter collection
  *
- * This function generates a BPF representation of the given filter DB.
+ * This function generates a BPF representation of the given filter collection.
  * Returns a pointer to a valid bpf_program on success, NULL on failure.
  *
  */
-struct bpf_program *gen_bpf_generate(const struct db_filter *db)
+struct bpf_program *gen_bpf_generate(const struct db_filter_col *col)
 {
 	int rc;
 	struct bpf_state state;
 
 	memset(&state, 0, sizeof(state));
-	state.arch = db->arch;
-	state.attr = &db->attr;
+	state.attr = &col->attr;
 
 	state.bpf = malloc(sizeof(*(state.bpf)));
 	if (state.bpf == NULL)
 		return NULL;
 	memset(state.bpf, 0, sizeof(*(state.bpf)));
 
-	rc = _gen_bpf_build_bpf(&state, db);
+	rc = _gen_bpf_build_bpf(&state, col);
 	if (rc < 0)
 		goto bpf_generate_end;
 
